/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDICONNECTION_H
#define MIDICONNECTION_H

#include <atomic>

#include "bindmidi.h"
#include "bindutil.h"
#include "boundarray.h"
#include "methodqueue.h"
#include "midievent.h"
#include "processclosure.h"
#include "source.h"
#include "statusserver.h"

namespace bipscript {
namespace midi {

template <class T> class MessageEventClosure : public ProcessClosure {
    T mmessage;
    const float mmeasure;

  protected:
    void addParameters() {
        MessageEvent *event = new MessageEvent(*(new T(mmessage)), mmeasure);
        BindUtil::PushRef(vm, event, binding::MessageEventObject);
    }

  public:
    MessageEventClosure(ScriptFunction function, T &mesg, RawEvent *evt)
        : ProcessClosure(function, evt->start()), mmessage(mesg), mmeasure(evt->measure()) {}
};

class Meter {
    bool mon;
    samples_t lasttime;
    static const samples_t REPORT_PERIOD = 11025; // 44100
  public:
    Meter() : mon(false) {}
    void update(void *node, uint16_t index, bool events, samples_t time) {
        StatusServer &statusserver = StatusServer::instance();
        if (events) {
            if (!mon) {
                statusserver.midiLevel(node, index, true);
                mon = true;
            }
            lasttime = time;
        } else if (mon && time - lasttime > REPORT_PERIOD) {
            statusserver.midiLevel(node, index, false);
            mon = false;
        }
    }
};

class Source;

class Connection {
    Source *msource;
    std::atomic<bool> handlerDefined;
    std::atomic<ScriptFunction *> onControlHandler;
    std::atomic<ScriptFunction *> onNoteOnHandler;
    std::atomic<ScriptFunction *> onNoteOffHandler;
    std::atomic<ScriptFunction *> onProgramChangeHandler;
    Meter mmeter;

  public:
    Connection(Source *source)
        : msource(source), handlerDefined(false), onControlHandler(0), onNoteOnHandler(0),
          onNoteOffHandler(0), onProgramChangeHandler(0) {}
    ~Connection() {}
    Source *source() { return msource; }
    // methods to access events in current period
    virtual RawEvent *firstEvent() = 0;
    virtual RawEvent *nextEvent() = 0;
    void onControl(ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            throw std::logic_error(
                "onControl handler should take a single Midi.MessageEvent argument");
        }
        onControlHandler.store(new ScriptFunction(handler));
        handlerDefined.store(true);
    }
    void onNoteOn(ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            throw std::logic_error(
                "onNoteOn handler should take a single Midi.MessageEvent argument");
        }
        onNoteOnHandler.store(new ScriptFunction(handler));
        handlerDefined.store(true);
    }
    void onNoteOff(ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            throw std::logic_error(
                "onNoteOff handler should take a single Midi.MessageEvent argument");
        }
        onNoteOffHandler.store(new ScriptFunction(handler));
        handlerDefined.store(true);
    }
    void onProgramChange(ScriptFunction &handler) {
        if (handler.numargs() != 2) {
            throw std::logic_error(
                "onProgramChange shandler should take a single Midi.MessageEvent argument");
        }
        onProgramChangeHandler.store(new ScriptFunction(handler));
        handlerDefined.store(true);
    }
    void fireEvents() {
        if (handlerDefined.load()) {
            MethodQueue &mq = MethodQueue::instance();
            ScriptFunction *ccHandler = onControlHandler.load();
            ScriptFunction *onHandler = onNoteOnHandler.load();
            ScriptFunction *offHandler = onNoteOffHandler.load();
            ScriptFunction *progHandler = onProgramChangeHandler.load();
            RawEvent *evt = firstEvent();
            while (evt) {
                RawMessage &mesg = evt->message();
                if (mesg.command() == Control::COMMAND && ccHandler) {
                    Control control(mesg.bytes());
                    mq.dispatch(new MessageEventClosure<Control>(*ccHandler, control, evt));
                } else if (mesg.command() == NoteOn::COMMAND && onHandler) {
                    NoteOn noteOn(mesg.bytes());
                    mq.dispatch(new MessageEventClosure<NoteOn>(*onHandler, noteOn, evt));
                } else if (mesg.command() == NoteOff::COMMAND && offHandler) {
                    NoteOff noteOff(mesg.bytes());
                    mq.dispatch(new MessageEventClosure<NoteOff>(*offHandler, noteOff, evt));
                } else if (mesg.command() == ProgramChange::COMMAND && progHandler) {
                    ProgramChange programChange(mesg.bytes());
                    mq.dispatch(
                        new MessageEventClosure<ProgramChange>(*progHandler, programChange, evt));
                }
                evt = nextEvent();
            }
        }
    }
    void restartScript() {
        ScriptFunctionCollector &collector = ScriptFunctionCollector::instance();
        ScriptFunction *ccHandler = onControlHandler.load();
        if (ccHandler) {
            collector.recycle(ccHandler);
            onControlHandler.store(nullptr);
        }
        ScriptFunction *onHandler = onNoteOnHandler.load();
        if (onHandler) {
            collector.recycle(onHandler);
            onNoteOnHandler.store(nullptr);
        }
        ScriptFunction *offHandler = onNoteOffHandler.load();
        if (offHandler) {
            collector.recycle(offHandler);
            onNoteOffHandler.store(nullptr);
        }
        ScriptFunction *progHandler = onProgramChangeHandler.load();
        if (progHandler) {
            collector.recycle(progHandler);
            onProgramChangeHandler.store(nullptr);
        }
        handlerDefined.store(false);
    }
    void reportLevels(void *node, uint16_t index, samples_t time) {
        mmeter.update(node, index, firstEvent() != nullptr, time);
    }
};

class Source : virtual public AbstractSource {
    BoundArray<Connection> bus;
    uint32_t mchannels;

  protected:
    Connection *midiOutput(uint32_t index) { return &bus[index]; }
    void midiOutput(Connection *connection, uint32_t index) {
        bus.set(index, connection, binding::ConnectionObject);
    }

  public:
    Source(uint32_t channels) : bus(channels), mchannels(channels) {}
    BoundArray<Connection> &midioutbus() { return bus; }
    HSQOBJECT midiout() { return bus.hsqobject(); }
    uint32_t channels() { return mchannels; }

    // event handlers
    void onControl(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onControl(handler);
        }
    }
    void onNoteOn(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onNoteOn(handler);
        }
    }
    void onNoteOff(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onNoteOff(handler);
        }
    }
    void onProgramChange(ScriptFunction &handler) {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].onProgramChange(handler);
        }
    }

  protected:
    void fireMidiEvents() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].fireEvents();
        }
    }
    void restartScript() {
        for (uint32_t i = 0; i < mchannels; i++) {
            bus[i].restartScript();
        }
    }
    void reportConnections(void *node) {
        void *array[mchannels];
        for (uint32_t i = 0; i < mchannels; i++) {
            array[i] = &bus[i];
        }
        StatusServer::instance().midiconnections(node, array, mchannels);
    }
    void reportLevels(void *node, samples_t time) {
        StatusServer &statusserver = StatusServer::instance();
        if (statusserver.active()) {
            for (uint32_t i = 0; i < mchannels; i++) {
                bus[i].reportLevels(node, i, time);
            }
        }
    }
};

} // namespace midi
} // namespace bipscript

#endif // MIDICONNECTION_H
