/*
 * This file is part of Bipscript.
 *
 * Bipscript is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bipscript is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bipscript.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MIDIMESSAGE_H
#define MIDIMESSAGE_H

#include "refcountable.h"
#include <stdint.h>

namespace bipscript {
namespace midi {

// events at the same measure value are sorted in this order:
enum Priority {
    REALTIME,
    TIMESIG,
    KEYSIG,
    TEMPO,
    METATEXT,
    PROGRAM,
    CONTROL,
    CHANNELOTHER,
    NOTEOFF,
    NOTEON
};

class Message : public RefCountable<Message> {
  public:
    virtual uint16_t size() const = 0;
    virtual void exportBytes(uint8_t *) = 0;
    virtual uint8_t command() = 0;
    virtual int priority() = 0;           // for sorting
    virtual bool meta() { return false; } // SMF meta event
    virtual ~Message() {}
};

class ChannelMessage : public Message {
  protected:
    const uint8_t mchannel; // 0-based
    static uint8_t parseChannel(uint8_t byte) { return byte & 0x0F; }

  public:
    ChannelMessage(uint8_t channel) : mchannel(channel) {}
    int priority() override { return Priority::CHANNELOTHER; }
    uint8_t channel() { return mchannel; }
    // default size, override if needed
    uint16_t size() const override { return 3; }
};

class Control : public ChannelMessage {
    const uint8_t mcontroller;
    const uint8_t mvalue;

  public:
    static const uint8_t COMMAND = 0xB0;
    uint8_t command() override { return COMMAND; }
    int priority() override { return Priority::CONTROL; }
    Control(uint8_t controller, uint8_t value, uint8_t channel = 0)
        : ChannelMessage(channel), mcontroller(controller), mvalue(value) {}
    Control(uint8_t *data)
        : ChannelMessage(parseChannel(data[0])), mcontroller(data[1]), mvalue(data[2]) {}
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = COMMAND + mchannel;
        buffer[1] = mcontroller;
        buffer[2] = mvalue;
    }
    uint8_t controller() { return mcontroller; }
    uint8_t value() { return mvalue; }
};

class NoteOn : public ChannelMessage {
    const uint8_t mpitch;
    const uint8_t mvelocity;

  public:
    static const uint8_t COMMAND = 0x90;
    uint8_t command() override { return COMMAND; }
    int priority() override { return Priority::NOTEON; }
    NoteOn(uint8_t pitch, uint8_t velocity, uint8_t channel = 0)
        : ChannelMessage(channel), mpitch(pitch), mvelocity(velocity) {}
    NoteOn(uint8_t *data)
        : ChannelMessage(parseChannel(data[0])), mpitch(data[1]), mvelocity(data[2]) {}
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = COMMAND + mchannel;
        buffer[1] = mpitch;
        buffer[2] = mvelocity;
    }
    uint8_t pitch() { return mpitch; }
    uint8_t velocity() { return mvelocity; }
};

class NoteOff : public ChannelMessage {
    const uint8_t mpitch;
    const uint8_t mvelocity;

  public:
    static const uint8_t COMMAND = 0x80;
    uint8_t command() override { return COMMAND; }
    int priority() override { return Priority::NOTEOFF; }
    NoteOff(uint8_t pitch, uint8_t velocity = 0, uint8_t channel = 0)
        : ChannelMessage(channel), mpitch(pitch), mvelocity(velocity) {}
    NoteOff(uint8_t *data)
        : ChannelMessage(parseChannel(data[0])), mpitch(data[1]), mvelocity(data[2]) {}
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = COMMAND + mchannel;
        buffer[1] = mpitch;
        buffer[2] = mvelocity;
    }
    uint8_t pitch() { return mpitch; }
    uint8_t velocity() { return mvelocity; }
};

class PitchBend : public ChannelMessage {
    const uint16_t mvalue; // 0 to 16,384, center = 8192

  public:
    static const uint8_t COMMAND = 0xE0;
    uint8_t command() override { return COMMAND; }
    PitchBend(uint16_t value, uint8_t channel = 0) : ChannelMessage(channel), mvalue(value) {}
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = COMMAND + mchannel;
        buffer[1] = mvalue & 0x7F; // LSB
        buffer[2] = mvalue >> 7;   // MSB
    }
    uint8_t value() { return mvalue; }
};

class ProgramChange : public ChannelMessage {
    const uint8_t mprogram;

  public:
    static const uint8_t COMMAND = 0xC0;
    uint8_t command() override { return COMMAND; }
    int priority() override { return Priority::PROGRAM; }
    ProgramChange(uint8_t program, uint8_t channel = 0)
        : ChannelMessage(channel), mprogram(program) {}
    ProgramChange(uint8_t *data) : ChannelMessage(parseChannel(data[0])), mprogram(data[1]) {}
    void exportBytes(uint8_t *buffer) override {
        buffer[0] = COMMAND + mchannel;
        buffer[1] = mprogram;
    }
    uint8_t program() { return mprogram; }
    uint16_t size() const override { return 2; }
};

class RealTimeMessage : public Message {
  public:
    int priority() override { return Priority::REALTIME; }
};

class ClockBeat : public RealTimeMessage {
  public:
    static const uint8_t COMMAND = 0xF8;
    uint8_t command() override { return COMMAND; }
    void exportBytes(uint8_t *buffer) override { buffer[0] = COMMAND; }
    uint16_t size() const override { return 1; }
};

class ClockStart : public RealTimeMessage {
  public:
    static const uint8_t COMMAND = 0xFA;
    uint8_t command() override { return COMMAND; }
    void exportBytes(uint8_t *buffer) override { buffer[0] = COMMAND; }
    uint16_t size() const override { return 1; }
};

class ClockContinue : public RealTimeMessage {
  public:
    static const uint8_t COMMAND = 0xFB;
    uint8_t command() override { return COMMAND; }
    void exportBytes(uint8_t *buffer) override { buffer[0] = COMMAND; }
    uint16_t size() const override { return 1; }
};

class ClockStop : public RealTimeMessage {
  public:
    static const uint8_t COMMAND = 0xFC;
    uint8_t command() override { return COMMAND; }
    void exportBytes(uint8_t *buffer) override { buffer[0] = COMMAND; }
    uint16_t size() const override { return 1; }
};

} // namespace midi
} // namespace bipscript

#endif // MIDIMESSAGE_H
